import pandas as pd
import uuid
import numpy as np
from dataclasses import dataclass
from typing import Union


def chooser(data: Union[pd.Series, pd.DataFrame], myproperty: str=''):
    """Function to choose elements from a pd.Series randomly, which consists of keys representing the elements and probabilities as values [-> Statistics object].

    Args:
        data (pd.Series | pd.DataFrame): input data to chose from which can be either a pandas.Series or a pandas.DataFrame
        myproperty (str, optional): If the data is in form of a pandas.DataFrame then the myproperty property defines the column to chose from
    Returns:
        _type_: randomly chosen element from pandas.Series or pandas.DataFrame
    """

    if not myproperty:
        data = pd.Series(data)
    else:
        # if property is nested
        types = data.keys()
        data = pd.Series(index=types,
                         data=[data[x][myproperty] for x in types])

    # take only probabilities which are greater than 0
    data = data[data > 0]

    # normalize probabilities to 0
    data /= (data.sum())

    # choose a random number between 0 and 1 from a uniform distribution
    u = np.random.uniform()

    # choose an element from the Series or DataFrame respectively randomly.
    choose = data[u < data.cumsum()].index[0]

    return choose


def duration_decorator(func):
    """Decorator function for duration.

    This decorator transforms duration from timedelta to total seconds, then performs the function on the
    seconds (e.g. fct_duration for generating durations from a probability distribution), afterwards  it transforms
    the output back to a pd.Timedelta object."""

    def wrapper(*args, **kwargs):

        args = map(lambda x: (pd.Timedelta(x)).total_seconds(), args)
        kwargs = {k: pd.Timedelta(v).total_seconds() for k, v in kwargs.items()}
        # kwargs = map(lambda x: (pd.Timedelta(x)).total_seconds(), kwargs)
        result = func(*args, **kwargs)
        result = to_timedelta(round(result))
        return result

    return wrapper


def normalize(pdf: Union[np.ndarray, pd.Series]) -> Union[np.ndarray, pd.Series]:
    """Normalize probability density function (pdf).

    Args:
        pdf (np.ndarray | pd.Series): probability density as numpy array or pandas.Series 

    Returns:
        np.ndarray | pd.Series: normalized probability density
    """
    pdf = pdf / np.sum(pdf)
    return pdf


def to_timedelta(time: Union[str, float, pd.Timedelta]) -> pd.Timedelta:
    """Transform a time in format string or float to a pandas.Timedelta object. 

    The function raises an exception if the format is not a string, float or pandas.Timedelta object

    Args:
        time (str | float | pd.Timedelta): time in the format string, float or already pandas.Timedelta

    Raises:
        Exception: _description_

    Returns:
        pd.Timedelta: _description_
    """

    if isinstance(time, str):
        value = pd.Timedelta(time)

    elif isinstance(time, (int, float)):
        value = pd.Timedelta(seconds=time)

    elif isinstance(time, pd.Timedelta):
        value = time

    else:
        raise Exception('Value is of unknown type:', type(value), '. It should be of type str, float, or pandas.Timedelta')

    return value


@dataclass
class Base:
    """Base class of pysimdeum for generating objects. 
    
    The only argument of this parent class is an id which can either be set or will be set to a unique identifier
    """

    id: str = str(uuid.uuid4())



import os
import toml
from dataclasses import dataclass, field
# from ..data.NL.end_uses.pattern.pat_dishwasher import dishwasher_daily_pattern, dishwasher_enduse_pattern
from pysimdeum.data.NL.end_uses.pattern.pat_dishwasher import dishwasher_daily_pattern, \
    dishwasher_enduse_pattern
from pysimdeum.data.NL.end_uses.pattern.pat_ktap import ktap_daily_pattern
from pysimdeum.data.NL.end_uses.pattern.pat_washing_machine import washingmachine_daily_pattern, \
    washingmachine_enduse_pattern
dirname = os.path.dirname(__file__)


@dataclass
class Statistics:
    """Statistics dataclass that contains all the relevant statistical information for pysimdeum."""

    country: str = 'NL'   
    household: dict = field(default_factory=dict)
    diurnal_pattern: dict = field(default_factory=dict)
    end_uses: dict = field(default_factory=dict)
    statisticsdir: str = os.path.join(dirname, '../data/NL/')  # TODO: Find good solution for this dirty statistics file workaround

    def __post_init__(self, country='NL'):
        self.country = country

        # Load household statistics
        household_file = os.path.join(self.statisticsdir, 'household_statistics.toml')
        self.household = toml.load(open(household_file, 'r'))

        # Load diurnal pattern statistics
        diurnal_pattern_file = os.path.join(self.statisticsdir, 'diurnal_patterns.toml')
        self.diurnal_pattern = toml.load(open(diurnal_pattern_file, 'r'))

        # load end-uses:
        self.end_uses = dict()
        path2end_use = os.path.join(self.statisticsdir, 'end_uses')

        bathtub_file = os.path.join(path2end_use, 'Bathtub.toml')
        brtap_file = os.path.join(path2end_use, 'BathroomTap.toml')
        dishwasher_file = os.path.join(path2end_use, 'Dishwasher.toml')
        kitchen_tap_file = os.path.join(path2end_use, 'KitchenTap.toml')
        outside_tap_file = os.path.join(path2end_use, 'OutsideTap.toml')
        shower_file = os.path.join(path2end_use, 'Shower.toml')
        washing_machine_file = os.path.join(path2end_use, 'WashingMachine.toml')
        wc_file = os.path.join(path2end_use, 'Wc.toml')

        self.end_uses['Wc'] = toml.load(open(wc_file, 'r'))
        self.end_uses['Bathtub'] = toml.load(open(bathtub_file, 'r'))
        self.end_uses['BathroomTap'] = toml.load(open(brtap_file, 'r'))
        self.end_uses['Dishwasher'] = toml.load(open(dishwasher_file, 'r'))
        self.end_uses['KitchenTap'] = toml.load(open(kitchen_tap_file, 'r'))
        self.end_uses['OutsideTap'] = toml.load(open(outside_tap_file, 'r'))
        self.end_uses['Shower'] = toml.load(open(shower_file, 'r'))
        self.end_uses['WashingMachine'] = toml.load(open(washing_machine_file, 'r'))

        # Pattern
        self.end_uses['WashingMachine']['daily_pattern'] = washingmachine_daily_pattern()
        self.end_uses['WashingMachine']['enduse_pattern'] = washingmachine_enduse_pattern()
        self.end_uses['Dishwasher']['daily_pattern'] = dishwasher_daily_pattern()
        self.end_uses['Dishwasher']['enduse_pattern'] = dishwasher_enduse_pattern()
        self.end_uses['KitchenTap']['daily_pattern'] = ktap_daily_pattern()


def main():

    stats = Statistics()


if __name__ == '__main__':

    main()


import pandas as pd
from typing import Union

from pysimdeum.core.house import House, Property
from pysimdeum.core.statistics import Statistics

def create_diurnal_pattern(statistics: Statistics) -> pd.Series:
    
    num_sim = 500
    time = pd.timedelta_range(start='00:00:00', end='23:59:59', freq='1S')
    diurnal_pattern = pd.Series(index=time).fillna(0)

    for i in range(num_sim):
        prop = Property(statistics=statistics)
        house = prop.built_house()
        house.populate_house()

        for user in house.users:
            presence = user.compute_presence(weekday=True, statistics=statistics)
            presence = presence.fillna(0)
            diurnal_pattern = diurnal_pattern + presence
    return diurnal_pattern

def create_usage_data(houses: Union[list, House]): #TODO I am not able to tell that it should be a list[str], list[House] or House
    if type(houses) == list:
        appliance_data = pd.DataFrame()
        total_water_usage = 0
        total_users = 0
        total_number_of_days = 0
        appliance_data['total'] = 0
        for inputp in houses:
            if type(inputp) == str: #input file list
                prop = Property()
                loadedhouse = prop.built_house(housefile=inputp)
            else:
                loadedhouse = inputp
            one_appliance_data, water_usage, users, number_of_days, patterns = _create_data(loadedhouse)
            for appliance in one_appliance_data.index.values:
                if appliance in appliance_data.index.values:
                    appliance_data.loc[appliance, 'total'] += one_appliance_data.loc[appliance, 'total']
                else:
                    appliance_data.loc[appliance, 'total'] = one_appliance_data.loc[appliance, 'total']
            total_water_usage += water_usage
            total_users += users
            total_number_of_days += number_of_days*patterns
        appliance_data['percentage'] = (appliance_data['total']/total_water_usage)*100
        appliance_data['pp'] = appliance_data['total']/total_users
        appliance_data['pppd'] = appliance_data['pp']/total_number_of_days

    elif type(houses) == House:
        appliance_data, total_water_usage, total_users, total_number_of_days, total_patterns = _create_data(houses)
        
    else:
        print("Error: input should either be a House or a list of Houses")
    
    return appliance_data, total_water_usage, total_users, total_number_of_days

def _create_data(inputproperty):
    total_water_usage = float(inputproperty.consumption.sum('user').sum('time').sum('enduse').sum('patterns').values)
    total_patterns = len(inputproperty.consumption.patterns)
    total_users = len(inputproperty.users)
    appliance_data = inputproperty.consumption.sum('user').sum('time').sum('patterns').to_dataframe('total')
    appliance_data['percentage'] = (appliance_data['total']/total_water_usage)*100
    appliance_data['pp'] = appliance_data['total']/total_users
    number_of_seconds = len(inputproperty.consumption)
    total_number_of_days = number_of_seconds/(60*60*24)
    appliance_data['pppd'] = (appliance_data['pp']/total_patterns)/total_number_of_days
    return appliance_data, total_water_usage, total_users, total_number_of_days, total_patterns


import numpy as np
import pandas as pd
import xarray as xr
import pickle
from datetime import datetime
from typing import Any, Union
from pysimdeum.core.utils import Base, chooser, normalize
from pysimdeum.core.statistics import Statistics
from pysimdeum.core.user import User
import pysimdeum.core.end_use as EndUses
from dataclasses import dataclass, field

# TODO: Implement multiple appliances which will be later on divided over the users, so they are not blocked
# TODO: Link data to OOPNET
# TODO: Increase speed of program
# TODO: Documentation
# TODO: Think about if Property class is necessary or if it can be integrated into the House class


@dataclass
class Property(Base):
    """Property represents a parcel of land on which a house is/can be build.

    Additionally, it contains information on statistics, and the location of the house, plus a connection to the oopnet_id
    """

    house_type: str = ""
    _house: Any = None
    statistics: Statistics = None
    country: str = "NL"

    # TODO: implement following quantities
    oopnet_id: str = ""
    lattitude: float = np.nan
    longitude: float = np.nan
    x_coordinate: float = np.nan
    y_coordinate: float = np.nan


    def __post__init__(self):
        if Statistics is None:
            self.statistics = Statistics(country=self.country)


    def _choose_type(self, statistics: Statistics=None) -> str:

        if not statistics:
            raise Exception('Statistics object has to be defined')
        else:
            self.house_type = chooser(data=statistics.household, myproperty='households')

        return self.house_type

    def built_house(self, house_type: str="", housefile: str=""):

        # TODO: House chooser seems not to work if API:built_house does not specify a house_type

        if housefile:
            with open(housefile, 'rb') as f:
                new_house = pickle.load(f)
            self.house = new_house
        else:
            if house_type:
                self.house_type = house_type
            else:
                self._choose_type(self.statistics)
            self.house = House(id=self.id,
                            house_type=self.house_type,
                            statistics=self.statistics,
                            oopnet_id=self.oopnet_id,
                            lattitude=self.lattitude,
                            longitude=self.longitude,
                            x_coordinate=self.x_coordinate,
                            y_coordinate=self.y_coordinate)

        return self.house


@dataclass
class House(Property):
    """pysimdeum House containting information on users, appliances and consumption.

    This class is a child of the `Property` class. It containts methods for populating and furnishing a house, and 
    methods to initialise and run simulations.
    """

    users: list = field(default_factory=list)  # List of users/inhabitants present in the house
    appliances: list = field(default_factory=list)  # List of appliances/water end-use devices in the house
    consumption: xr.DataArray = field(default_factory=xr.DataArray)  # property to store the consumption of a house

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}:\n\tid\t=\t{self.id}\n\ttype\t=' \
               f'\t{self.house_type}\n\tuser\t=\t{len(self.users)}\n' \
               f'\tappliances\t=\t{list(map(lambda x: x.__class__.__name__, self.appliances))}'
    
    def __str__(self):
        return self.__repr__()

    def populate_house(self) -> None:

        # job statistic
        job_stats = normalize(pd.Series(self.statistics.household[self.house_type]['job']))

        # division age statistics
        age_stats = normalize(pd.Series(self.statistics.household[self.house_type]['division_age']))

        # division gender statistics
        gender_stats = normalize(pd.Series(self.statistics.household[self.house_type]['division_gender']))

        if self.house_type == 'one_person':

            age = chooser(data=age_stats)
            gender = chooser(data=gender_stats)
            job = False

            if age == 'adult':
                u = np.random.uniform()
                if u < job_stats[gender]:
                    job = True

            self.users = [User(id='user_1', age=age, gender=gender, job=job)]

        elif self.house_type == 'two_person':
            # todo: implement same sex households

            # choose age
            age1 = chooser(data=age_stats)
            age2 = chooser(data=age_stats)

            # choose gender
            gender = chooser(data=gender_stats)
            gender1, gender2 = gender.split('_')

            # choose job
            job1 = False
            job2 = False
            u = np.random.uniform()

            # have a job
            if age1 == 'senior':
                if age2 == 'adult':
                    if u < job_stats['only_female'] / (job_stats['only_female'] + job_stats['neither_person']):
                        job2 = True

            if age2 == 'senior':
                if age1 == 'adult':
                    if u < job_stats['only_male'] / (job_stats['only_male'] + job_stats['neither_person']):
                        job1 = True

            if (age1 == 'adult') and (age2 == 'adult'):
                job = chooser(job_stats)
                if job == 'both':
                    job1 = True
                    job2 = True
                elif job == 'only_male':
                    job1 = True
                    job2 = False
                elif job == 'only_female':
                    job1 = False
                    job2 = True
                elif job == 'neither_person':
                    job1 = False
                    job2 = False
                else:
                    raise Exception('Unknown job attribute, not implemented.')

            self.users = [User(id='user_1', age=age1, gender=gender1, job=job1),
                          User(id='user_2', age=age2, gender=gender2, job=job2)]

        elif self.house_type == 'family':

            averagenumpeople = self.statistics.household[self.house_type]['people']
            minnum = 2
            maxnum = 5
            rNum = np.random.binomial(n=maxnum - minnum, p=(averagenumpeople - minnum) / (maxnum - minnum)) + minnum

            if rNum == 2:  # mother and child/teen

                # u = pm.Uniform.dist().random()
                u = np.random.uniform()

                # mother
                job = False
                if u < job_stats[['both', 'only_female']].sum():  # not correct, this is for two partners (comment
                    # from Mirjam Blokker)
                    job = True
                mother = User(id='user_1', age='adult', job=job, gender='female')

                # child
                gender = chooser(gender_stats)
                age = chooser(age_stats[['child', 'teen']])
                child = User(id='user_2', age=age, job=False, gender=gender)

                self.users = [mother, child]

            elif rNum in [3, 4, 5]:

                # Generate parents
                job = chooser(job_stats)
                if job == 'both':
                    f_job = True
                    m_job = True
                elif job == 'only_male':
                    f_job = True
                    m_job = False
                elif job == 'only_female':
                    f_job = False
                    m_job = True
                elif job == 'neither_person':
                    f_job = False
                    m_job = False

                family = [User(id='user_1', gender='male', age='adult', job=f_job),  # father
                          User(id='user_2', gender='female', age='adult', job=m_job)]  # mother

                # add child/teen until family size is reached
                for numchild in range(2, rNum):
                    gender = chooser(gender_stats)
                    age = chooser(age_stats[['child', 'teen']])
                    family += [User(id='user_' + str(numchild+1), gender=gender, age=age, job=False)]  #
                    # additional
                    # child

                self.users = family

            else:
                raise Exception('Family size exceeded!')
        else:

            raise NotImplementedError('Household type is not implemented')

    def furnish_house(self):

        for key, appliances in self.statistics.end_uses.items():

            penetration = appliances['penetration']
            classname = appliances['classname']
            inhabitants = str(len(self.users))

            u = np.random.uniform() * 100  # probability in percent

            # penetration dependent on number of inhabitants
            if isinstance(penetration, dict):
                penetration = penetration[inhabitants]

            if u <= penetration:
                if classname == 'Shower':
                    showertype = chooser(appliances['subtype'], 'penetration')
                    eu_instance = getattr(EndUses, showertype)(statistics=appliances)
                elif classname == 'Wc':
                    wctype = chooser(appliances['subtype'], 'penetration')
                    eu_instance = getattr(EndUses, wctype)(statistics=appliances)
                else:
                    eu_instance = getattr(EndUses, classname)(statistics=appliances)
                self.appliances.append(eu_instance)

    def init_consumption(self):
        # todo: can get rid off, functionality shifted to simulate

        time = pd.TimedeltaIndex(start='00:00:00', end='24:00:00', freq='1s', closed='left')
        users = [x.id for x in self.users] + ['household']
        enduse = [x.name for x in self.appliances]

        self.consumption = xr.DataArray(data=np.zeros((len(time), len(users), len(enduse))),
                                        coords=[time, users, enduse],
                                        dims=['time', 'user', 'enduse'])
        return self.consumption

    def simulate(self, date=None, duration='1 day', num_patterns=1):

        if date is None:
            date = datetime.now().date()
        try:
            timedelta = pd.to_timedelta(duration)
        except:
            print('Warning: duration unrecognized defaulted to 1 day')
            timedelta = pd.to_timedelta('1 day')
        # time = pd.timedelta_range(start='00:00:00', end='24:00:00', freq='1s', closed='left')
        # time = pd.date_range(start=date, end=date + timedelta, freq='1s', closed='left')
        time = pd.date_range(start=date, end=date + timedelta, freq='1s')
        users = [x.id for x in self.users] + ['household']
        enduse = [x.statistics['classname'] for x in self.appliances]
        patterns = [x for x in range(0, num_patterns)]
        consumption = np.zeros((len(time), len(users), len(enduse), num_patterns))

        for num in patterns:
            for k, appliance in enumerate(self.appliances):

                consumption = appliance.simulate(consumption, users=self.users, ind_enduse=k, pattern_num=num)

        self.consumption = xr.DataArray(data=consumption, coords=[time, users, enduse, patterns], dims=['time', 'user', 'enduse', 'patterns'])

        return self.consumption

    def save_house(self, outputname):
#        if self.consumption == None: #only save simulated houses
#            self.simulate()
        with open(outputname + '.house', 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
    
        
# this class is introduced for storage purposes. 
# It is meant for those cases where only demand data per house is needed 
# and not the individual data of the users and or appliances
# the conumption datarray therefore only contains totals
# user and appliance data is removed for now
@dataclass
class HousePattern:

    house: Union[House, str]
    users: list = field(default_factory=list, init=False)
    appliances: list = field(default_factory=list, init=False)
    consumption: list = field(default_factory=list, init=False)

    def __post__init__(self, house):

        if type(self.house) == House:
            self.users = self.house.users
            self.appliances = self.house.appliances 
            self.consumption = self.house.consumption.sum('user').sum('enduse')
        
        elif type(self.house) == str:
            with open(house, 'rb') as f:
                new_house_pattern = pickle.load(f)
                self.users = new_house_pattern.users
                self.appliances = new_house_pattern.appliances
                self.consumption = new_house_pattern.consumption
   
    def save_house_pattern(self, outputname):
        with open(outputname + '.housepattern', 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
    



import matplotlib.pyplot as plt
import numpy as np
from typing import Union
from pysimdeum.core.statistics import Statistics

from pysimdeum.tools.helper import create_diurnal_pattern, create_usage_data
from pysimdeum.core.house import House

def plot_water_use_distribution(inputproperty: Union[list, House], plotsubject: str='percentage'):
    """Function to plot water use distribution betwee the different appliances as a pie graph [-> Axessubplot object]
    
    Args:
        inputproperty (list[House] | list[str] | House): the house or houses (either as a list of objects or paths to files) to be plotted
        plotsubject (str, optional): unit of plot, percentage, per person per day or per person default is percentage
    Returns:
        ax1 (matplotlib axessubplot): an matplotlib axes containing the plot (use plt.show() to render)
    """
    appliance_data, total_water_usage, total_users, total_number_of_days = create_usage_data(inputproperty)
    def func(pct, allvals):
        absolute = pct/100.*np.sum(allvals)
        return "{:.1f}L\n({:.1f}%)".format(absolute, pct)
    if plotsubject != 'total' and plotsubject != 'percentage' and plotsubject != 'pp' and plotsubject != 'pppd':
        print('Plotting subject not known. options are total, percentage (default), pppd (per person per day) and pp (per person)')
    else:
        labels = appliance_data.index.values
        sizes = appliance_data[plotsubject].values
        fig1, ax1 = plt.subplots()
        if plotsubject == 'percentage':
            ax1.pie(sizes, labels=labels, startangle=90, autopct='%1.1f%%')
        else:
            ax1.pie(sizes, labels=labels, startangle=90, autopct=lambda pct: func(pct, sizes))
        fig1.suptitle(plotsubject)
        ax1.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.
        return ax1

def createQcfdplot(houses: House):
    """Function to plot a cumulative flow plot [-> Axessubplot object]
    
    Args:
        houses (House): the house object to be plotted (other options not implemented yet)
    Returns:
        ax (matplotlib axessubplot): an matplotlib axes containing the plot (use plt.show() to render)
    """
    
    if type(houses) == House:
        n_bins = 100
        fig, ax = plt.subplots()
        x = houses.consumption.sel(patterns=0).sum('user').sum('enduse').values
        n, bins, patches = ax.hist(x, n_bins, density=True, histtype='step',
                            cumulative=True, label='Empirical')

        return ax
    else:
        # list of housefiles not implemented yet
        return None

def plot_demand(houses: House):
    """Function to plot demand. It will give four plots containing demand per user, per enduse, total of pattern 1 and all patterns/num patterns [-> figure object]
    
    Args:
        houses (House): the house object to be plotted (other options not implemented yet)
    Returns:
        fig (matplotlib figure): an matplotlib figure containing the plot (use plt.show() to render)
    """

    if type(houses) == House:
        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2,2, sharey=True)
        for user in houses.consumption.user.values:
            ax1.plot(houses.consumption['time'].values, houses.consumption.sel(user=user, patterns=0).sum('enduse').values, label=user)
        ax3.plot(houses.consumption['time'].values, houses.consumption.sel(patterns=0).sum('user').sum('enduse').values, label='total')
        for enduse in houses.consumption.enduse.values:
            ax2.plot(houses.consumption['time'].values, houses.consumption.sel(enduse=enduse, patterns=0).sum('user').values, label=enduse)
        ax4.plot(houses.consumption['time'].values, houses.consumption.sum('user').sum('enduse').sum('patterns').values/len(houses.consumption.patterns), label='average')
        ax1.legend()
        ax1.set_xlabel('time')
        ax1.set_ylabel('demand (l/s)')
        ax2.set_xlabel('time')
        ax3.legend()
        ax3.set_xlabel('time')
        ax3.set_ylabel('demand (l/s)')
        ax4.set_xlabel('time')
        ax2.legend()
        ax4.legend()
        return fig
    else:
        #list of housefiles not implemented yet
        test=2
        return None

def view_statistics(statistics: Statistics):
    """Function to plot an overview of the household statistics percentage one, two and family households for instance [-> Axessubplot object]
    
    Args:
        statistics (Statistics): the statistics object to be plotted
    Returns:
        ax (matplotlib axessubplot): an matplotlib axes containing the plot (use plt.show() to render)
    """
    
    data = []
    text = []
    for key in statistics.household.keys():
        temptext = []
        temptext.append(key + ': ' + str(statistics.household[key]['households']) + '\n')
        data.append(statistics.household[key]['households'])
        for key2 in statistics.household[key]['division_gender'].keys():
            temptext.append(key2 + ': ' + str(statistics.household[key]['division_gender'][key2]) + ' ')
        temptext.append('\n')
        for key3 in statistics.household[key]['division_age'].keys():
            temptext.append(key3 + ': ' + str(statistics.household[key]['division_age'][key3]) + ' ')
        temptext.append('\n')
        for key4 in statistics.household[key]['job'].keys():
            temptext.append(key4 + ': ' + str(statistics.household[key]['job'][key4]) + ' ')
        temptext.append('\n')
        text.append(''.join(temptext))

    ax = __create_pie_fig(data, text, 'household statistics and home presence settings')

    # home presence table
    rows = []
    cell_text= []
    for key in statistics.diurnal_pattern['child'].keys():
        for key2 in statistics.diurnal_pattern['child'][key].keys():
            rows.append(key + '-' + key2)
    columns= list(statistics.diurnal_pattern.keys())
    for key in statistics.diurnal_pattern['child'].keys():
        for key2 in statistics.diurnal_pattern['child'][key].keys():
            temp_text = []
            width = []
            for column in columns:
                temp_text.append(statistics.diurnal_pattern[column][key][key2])
                width.append(0.4)
            cell_text.append(temp_text)
    the_table = plt.table(cellText=cell_text, cellLoc='center', rowLabels=rows, colLabels=columns, colWidths=width,
                      loc='bottom')
    plt.subplots_adjust(bottom=0.45)
    
    return ax

def plot_diurnal_pattern(statistics: Statistics):
    """Function to plot a diurnal pattern (household presence) based on statistics [-> Axessubplot object]
    
    Args:
        statistics (Statistics): the Statistics object to be plotted
    Returns:
        ax (matplotlib axessubplot): an matplotlib axes containing the plot (use plt.show() to render)
    """
    
    diurnal_pattern = create_diurnal_pattern(statistics)
    fig, ax1 = plt.subplots()

    ax1.plot(diurnal_pattern.index, diurnal_pattern.values)
    
    return ax1

def __create_pie_fig(data, text, title):
    fig, ax = plt.subplots(figsize=(6, 3), subplot_kw=dict(aspect="equal"))


    wedges, texts = ax.pie(data, wedgeprops=dict(width=0.7), startangle=-80)

    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
    kw = dict(arrowprops=dict(arrowstyle="-"),
            bbox=bbox_props, zorder=0, va="center")

    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1)/2. + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        ax.annotate(text[i], xy=(x, y), xytext=(1.35*np.sign(x), 1.4*y),
                    horizontalalignment=horizontalalignment, **kw)

    ax.set_title(title)

    return ax




from pysimdeum.core.statistics import Statistics
from pysimdeum.core.house import Property, HousePattern, House


def built_house(house_type: str = "") -> House:

    stats = Statistics()
    prop = Property(statistics=stats)
    house = prop.built_house(house_type=house_type)
    house.populate_house()
    house.furnish_house()
    for user in house.users:
        user.compute_presence(statistics=stats)
    house.simulate(num_patterns=100)

    return house


import os
import pandas as pd


def ktap_daily_pattern(filename=None, resolution='1s'):

    if filename is None:

        pathname = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(pathname, 'dp_kitchen_tap.txt')

    data = pd.read_csv(filename, header=None).squeeze().values
    index = pd.timedelta_range(start='00:00:00', freq='15Min', end='24:00:00')
    s = pd.Series(data=data, index=index)
    s = s.resample(resolution).mean().interpolate(method='linear')
    # print(s.shape)
    s = s[s.index.days == s.index[0].days]
    # print(s.shape)
    return s



import pandas as pd
from datetime import datetime
from typing import Union

from pysimdeum.tools.helper import create_usage_data
from pysimdeum.core.house import HousePattern, Property, House

def export_water_use_distribution(inputproperty: Union[list, House], name: str='ApplianceWaterUse.xlsx'):
    appliance_data, total_water_usage, total_users, total_number_of_days = create_usage_data(inputproperty)
    writer = pd.ExcelWriter(name, engine = 'xlsxwriter')
    metadata = pd.DataFrame(index=['info'])
    metadata['Total number of Users'] = total_users
    metadata['Total water consumption'] = total_water_usage
    metadata['Total number of days'] = total_number_of_days
    metadata['Calculation date'] = datetime.now().date()
    appliance_data.to_excel(writer, sheet_name = 'data')
    metadata.to_excel(writer, sheet_name = 'metadata')
    writer.close()

def write_simdeum_patterns_to_ddg(houses: list, timestep: int, Q_option: str, patternfile_option: int, output_file: str):
    # house can be either be a list of filenames or a list of houses
    # timestep the output timestep of the pattern. minimum is 1 minute
    # Q_option for now only 'm3/h' TODO is this true? are the units not L/s?????
    # for now only 1 all patterns in 1 file
    # output_file file to write to

    output = __get_output_dataframe(houses, timestep) 
    test = 2
    

def write_simdeum_patterns_to_xlsx(houses: list, timestep: int, Q_option: str, patternfile_option: int, output_file: str):
    # after writeSimdeumPatternToXls (Matlab)
    # house can eithwer be a list of filenames or a list of houses
    # timestep the output timestep of the pattern
    # Q_option for now only 'm3/h' TODO is this true? are the units not L/s?????
    # for now only 1 all patterns in 1 file
    # output_file file to write to
    
    output = __get_output_dataframe(houses, timestep)
    output.to_excel(output_file)

def __get_output_dataframe(houses, timestep):
    if type(houses[0]) == str:
        count = 0
        output = pd.DataFrame()
        if '.housepattern' in houses[0]: #it are housepattern files
            for housepattern in houses:
                loadedhousepattern = HousePattern(housepattern)
                __get_housepattern_output(count, output, loadedhousepattern)
                count +=1
        else: #assumed to be house files
            for house in houses:
                prop = Property()
                loadedhouse = prop.built_house(housefile=house)
                __get_house_output(count, output, loadedhouse)
                count +=1
        summedoutput = pd.DataFrame()
        summedoutput['date'] = output['date'].values[0::timestep]
        for column in output.columns:
            if column != 'date':
                summedoutput[column] = output[column].groupby(output.index // timestep).sum().values
        return summedoutput

    else: #.houses
        count = 0
        output = pd.DataFrame()
        for house in houses:
            __get_house_output(count, output, house)
            count += 1
        summedoutput = pd.DataFrame()
        summedoutput['date'] = output['date'].values[0::timestep]
        for column in output.columns:
            if column != 'date':
                summedoutput[column] = output[column].groupby(output.index // timestep).sum().values
        return summedoutput

def __get_house_output(count, output, loadedhouse):
    if count == 0:
        datelist = []
        valueslist = []
        for i in range(0, len(loadedhouse.consumption.patterns)):
            datelist.extend(loadedhouse.consumption['time'].values)
            valueslist.extend(loadedhouse.consumption.sel(patterns=i).sum('user').sum('enduse').values)
        output['date'] = datelist
    else:
        valueslist = []
        for i in range(0, len(loadedhouse.consumption.patterns)):
            valueslist.extend(loadedhouse.consumption.sel(patterns=i).sum('user').sum('enduse').values)
    output['pysimdeum ' + str(count)] = valueslist

def __get_housepattern_output(count, output, loadedhousepattern):
    if count == 0:
        datelist = []
        valueslist = []
        for i in range(0, len(loadedhousepattern.consumption.patterns)):
            datelist.extend(loadedhousepattern.consumption['time'].values)
            valueslist.extend(loadedhousepattern.consumption.sel(patterns=i).values)
        output['date'] = datelist
    else:
        valueslist = []
        for i in range(0, len(loadedhousepattern.consumption.patterns)):
            valueslist.extend(loadedhousepattern.consumption.sel(patterns=i).values)
    output['pysimdeum ' + str(count)] = valueslist


from traits.api import HasStrictTraits, Either, Instance, Str, Any
import copy
import pandas as pd
import numpy as np
from utils import chooser, duration_decorator, normalize, to_timedelta


class EndUse(HasStrictTraits):
    """Base class for end-uses.

    """

    name = Str  # ... name of the end-use
    statistics = Any  # ... statistic object associated with end-use

    def __init__(self, name: str = 'EndUse', statistics=None):
        """Initialisation funciton of end-use class.

        Args:
            name: name of the specific end-use as string
            statistics: statistics object associated with end-use
        """
        super(EndUse, self).__init__()
        self.name = name
        self.statistics = statistics

    def init_consumption(self, users=None, time_resolution='1s'):
        """Initialization of a pandas dataframe to store the  consumptions.

        Args:
            users:  list with users
            time_resolution:  string with desired time resolution as python pandas `DateOffset` object
            (https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects)

        Returns:
            consumption as pandas `DataFrame` filled with zeros
        """

        if users:
            # produce datetime index
            index = pd.TimedeltaIndex(start='00:00:00', end='24:00:00', freq=time_resolution, closed='left')

            # name columns by users
            columns = ['user_' + str(x+1) for x, user in enumerate(users)]

            # initialise consumption dataframe with timedelta index and user columnnames, name it according to end-use
            # device and fill it with zeros.
            consumption = pd.DataFrame(data=0, index=index, columns=columns)
            consumption.name = self.name
        else:
            # raise an error if no users are defined.
            raise Exception('No Users are defined!')

        return consumption

    @staticmethod
    def usage_probability(time_resolution='1s'):
        """Produces uninformed prior.

        For more specific usage probabilities (washing machine, kitchen tap, dishwasher) overload this function by
        loading a usage pattern into it.
        """
        # produce datetime index
        index = pd.timedelta_range(start='00:00:00', end='24:00:00', freq=time_resolution, closed='left')

        prob = pd.Series(data=1, index=index)  # ... uniform probability over time and cast it into pandas series.
        prob /= prob.sum()  # ... normalization of the probabilities

        return prob

    def fct_frequency(self):
        """Placeholder for specific frequency probability function defined in specific EndUse"""

        raise NotImplementedError('Frequency function is not implemented yet!')

    def fct_duration(self):
        """Placeholder for specific duration probability function defined in specific EndUse"""

        raise NotImplementedError('Duration function is not implemented yet!')

    def fct_intensity(self):
        """Placeholder for specific intensity probability function defined in specific EndUse"""

        raise NotImplementedError('Intensity function is not implemented yet!')

    def fct_duration_intensity(self):
        """Computing duration and intensity for enduse"""

        duration = self.fct_duration()
        intensity = self.fct_intensity()

        return duration, intensity


class Bathtub(EndUse):
    """Class for Bathtub end-use."""


    def __init__(self, name='Bathtub', **kwargs):
        """Initialisation function of Bathtub end-use class.

        Args:
            name: End-use name as string.
            **kwargs: keyword arguments for super classes.
        """
        super(Bathtub, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self, age=None):
        """Random function computing the frequency of use for the Bathtub end-use class.

        Args:
            age: age of the user in years.

        Returns:
            distribution function from `numpy.random` to compute frequency of use.

        """
        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())
        average = f_stats['average'][age]

        return distribution(average)

    def fct_duration(self):
        """Function to compute the duration of Bathtub end-use.

        Comment: The duration is fixed to 10 minutes in this case.

        Returns:
            duration (fixed value as integer)

        """
        # fixed duration
        return int(to_timedelta(self.statistics['duration']).total_seconds())

    def fct_intensity(self):
        """Compute the intensity of Bathtub end-use.

        Returns:
            intensity (fixed value as float)

        """
        # fixed intensity
        return self.statistics['intensity']

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = self.usage_probability().values

        for j, user in enumerate(users):
            freq = self.fct_frequency(age=user.age)
            prob_user = user.presence.values

            for i in range(freq):

                duration, intensity = self.fct_duration_intensity()
                prob_joint = normalize(prob_user * prob_usage)

                u = np.random.uniform()
                start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
                end = start + duration

                consumption[start:end, j, ind_enduse] = intensity

                # block uses of current user for other enduses (this is possible because Python takes the object
                # itself and not a copy of the presence attribute) So no user can use two endtypes at the same time
                prob_user[start:end] = 0
                prob_user = normalize(prob_user)

                # block for other users (only possible because of fixed duration time)
                prob_usage[start-duration:end] = 0
                prob_usage = normalize(prob_usage)

        return consumption


class BathroomTap(EndUse):
    
    def __init__(self, name='BathroomTap', **kwargs):
        super(BathroomTap, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self):

        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())
        average = f_stats['average']
        return distribution(average)

    def fct_duration_intensity(self):

        subtype = chooser(self.statistics['subtype'], 'penetration')

        d_stats = self.statistics['subtype'][subtype]['duration']
        i_stats = self.statistics['subtype'][subtype]['intensity']

        dist = duration_decorator(getattr(np.random, d_stats['distribution'].lower()))
        mean = to_timedelta(np.log(to_timedelta(d_stats['average']).total_seconds()) - 0.5)
        duration = dist(mean=mean).total_seconds()

        dist = getattr(np.random, i_stats['distribution'].lower())
        low = i_stats['low']
        high = i_stats['high']

        intensity = dist(low=low, high=high)

        return duration, intensity

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = self.usage_probability().values

        for j, user in enumerate(users):
            freq = self.fct_frequency()
            prob_user = user.presence.values

            for i in range(freq):
                check = False

                duration, intensity = self.fct_duration_intensity()

                prob_joint = normalize(prob_user * prob_usage)

                # get rid of overlaps
                for tries in range(10):
                    u = np.random.uniform()
                    start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
                    end = int(start + duration)
                    if np.sum(consumption[start:end, j, ind_enduse]) == 0:
                        check = True
                        break

                if check:
                    prob_user[start:end] = 0
                    prob_user = normalize(prob_user)

                    prob_usage[start:end] = 0
                    prob_usage = normalize(prob_usage)

                    consumption[start:end, j, ind_enduse] = intensity

        return consumption


class Dishwasher(EndUse):

    def __init__(self, name='Dishwasher', **kwargs):
        super(Dishwasher, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self, numusers=None):

        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())

        df = pd.Series(f_stats['average'])
        average = df[str(numusers)]  * numusers
        return distribution(average)

    def fct_duration_pattern(self, start=None):
        pattern = self.statistics['enduse_pattern']
        # duration = pattern.index[-1] - pattern.index[0]
        return pattern

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = copy.deepcopy(self.statistics['daily_pattern'].values)
        freq = self.fct_frequency(numusers=len(users))

        for j, user in enumerate(users):
            if j == 0:
                prob_user = copy.deepcopy(user.presence)
            else:
                prob_user += user.presence

        prob_user = normalize(prob_user.values)
        j = len(users)

        prob_joint = normalize(prob_user * prob_usage)

        pattern = self.fct_duration_pattern().values
        duration = len(pattern)

        for i in range(freq):

            u = np.random.random()
            start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
            end = start + duration

            if end > (24 * 60 * 60):
                end = 24 * 60 * 60
            difference = end - start
            consumption[start:end, j, ind_enduse] = pattern[:difference]

            prob_user[start:end] = 0
            prob_user = normalize(prob_user)

            prob_usage[start - duration:end] = 0
            prob_usage = normalize(prob_usage)

        return consumption


class KitchenTap(EndUse):

    def __init__(self, name='KitchenTap', **kwargs):
        super(KitchenTap, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self, numusers=None):

        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())

        df = pd.Series(f_stats['average'])
        average = df[str(numusers)]

        df = pd.Series(f_stats['sigma'])
        sigma = df[str(numusers)]

        # Todo: find out which implementation is right? Mirjam (implemented here) or Wikipedia

        # Implementation according to Wikipedia
        # p = (sigma ** 2 - average) / (sigma ** 2)
        # r = (average ** 2) / (sigma ** 2 - average)

        # implementation according to Mirjam
        p = average / sigma ** 2
        r = p * average / (1 - p)

        return distribution(r, p)

    def fct_duration_intensity(self):

        subtype = chooser(self.statistics['subtype'], 'penetration')

        d_stats = self.statistics['subtype'][subtype]['duration']
        i_stats = self.statistics['subtype'][subtype]['intensity']

        dist = getattr(np.random, d_stats['distribution'].lower())
        mean = np.log(pd.Timedelta(d_stats['average']).total_seconds()) - 0.5

        duration = int(pd.Timedelta(seconds=dist(mean=mean)).total_seconds())

        dist = getattr(np.random, i_stats['distribution'].lower())
        low = i_stats['low']
        high = i_stats['high']

        intensity = dist(low=low, high=high)

        return duration, intensity

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = copy.deepcopy(self.statistics['daily_pattern'].values)

        for j, user in enumerate(users):
            if j == 0:
                prob_user = copy.deepcopy(user.presence)
            else:
                prob_user += user.presence

        prob_user = normalize(prob_user).values

        j = len(users)

        freq = self.fct_frequency(numusers=len(users))

        for i in range(freq):

            duration, intensity = self.fct_duration_intensity()

            prob_joint = normalize(prob_user * prob_usage)

            for tries in range(10):
                u = np.random.uniform()
                start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
                end = start + duration

                if np.sum(consumption[start:end, j, ind_enduse]) == 0:
                    check = True
                    break

            if check:
                prob_user[start:end] = 0
                prob_user = normalize(prob_user)

                prob_usage[start:end] = 0
                prob_usage = normalize(prob_usage)

                consumption[start:end, j, ind_enduse] = intensity

        return consumption


class OutsideTap(EndUse):

    def __init__(self, name='OutsideTap', **kwargs):
        super(OutsideTap, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self):

        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())
        average = f_stats['average']
        return distribution(average)

    def fct_duration_intensity(self):

        subtype = chooser(self.statistics['subtype'], 'penetration')

        d_stats = self.statistics['subtype'][subtype]['duration']
        i_stats = self.statistics['subtype'][subtype]['intensity']

        dist = getattr(np.random, d_stats['distribution'].lower())
        mean = np.log(pd.Timedelta(d_stats['average']).total_seconds()) - 0.5

        duration = int(pd.Timedelta(seconds=dist(mean=mean)).total_seconds())

        dist = getattr(np.random, i_stats['distribution'].lower())
        low = i_stats['low']
        high = i_stats['high']

        intensity = dist(low=low, high=high)

        return duration, intensity

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = self.usage_probability().values

        freq = 0
        for j, user in enumerate(users):
            if j == 0:
                prob_user = copy.deepcopy(user.presence)
            else:
                prob_user += user.presence
            freq += self.fct_frequency()

        prob_user = normalize(prob_user).values

        j = len(users)

        for i in range(freq):

            duration, intensity = self.fct_duration_intensity()

            prob_joint = normalize(prob_user * prob_usage)

            for tries in range(10):
                u = np.random.uniform()
                start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
                end = start + duration

                if np.sum(consumption[start:end, j, ind_enduse]) == 0:
                    check = True
                    break
                else:
                    print('consumption overlap')

            if check:
                prob_user[start:end] = 0
                prob_user = normalize(prob_user)

                prob_usage[start:end] = 0
                prob_usage = normalize(prob_usage)

                consumption[start:end, j, ind_enduse] = intensity

        return consumption


class Shower(EndUse):

    def __init__(self, name='Shower',**kwargs):
        super(Shower, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self, age=None):

        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())
        n = f_stats['n']
        p = f_stats['p'][age]

        return distribution(n, p)

    def fct_duration_intensity(self, age=None):

        d_stats = self.statistics['duration']
        distribution = getattr(np.random, d_stats['distribution'].lower())
        df = to_timedelta(d_stats['df'][age])

        df = int(df.total_seconds() / 60)
        duration = round(distribution(df))
        duration = int(pd.Timedelta(minutes=duration).total_seconds())

        intensity = self.statistics['subtype'][self.name]['intensity']

        return duration, intensity

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = self.usage_probability().values

        for j, user in enumerate(users):
            freq = self.fct_frequency(age=user.age)
            prob_user = user.presence.values

            for i in range(freq):
                check = False

                duration, intensity = self.fct_duration_intensity(age=user.age)

                prob_joint = normalize(prob_user * prob_usage)

                # get rid of overlaps
                for tries in range(10):
                    u = np.random.uniform()
                    start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
                    end = start + duration

                    if np.sum(consumption[start:end, j, ind_enduse]) == 0:
                        check = True
                        break
                    else:
                        print('consumption overlap')
                if check:
                    prob_user[start:end] = 0
                    prob_user = normalize(prob_user)

                    prob_usage[start:end] = 0
                    prob_usage = normalize(prob_usage)

                    consumption[start:end, j, ind_enduse] = intensity

        return consumption


class NormalShower(Shower):

    def __init__(self, name='NormalShower', **kwargs):
        super(NormalShower, self).__init__(**kwargs)
        self.name = name


class FancyShower(Shower):

    def __init__(self, name='FancyShower', **kwargs):
        super(FancyShower, self).__init__(**kwargs)
        self.name = name


class WashingMachine(EndUse):

    def __init__(self, name='WashingMachine', **kwargs):
        super(WashingMachine, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self, numusers=None):

        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())

        df = pd.Series(f_stats['average'])
        average = df[str(numusers)] * numusers
        return distribution(average)

    def fct_duration_pattern(self, start=None):
        pattern = self.statistics['enduse_pattern']
        # duration = pattern.index[-1] - pattern.index[0]
        return pattern

    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = copy.deepcopy(self.statistics['daily_pattern'].values)

        # for j, user in enumerate(users):
        freq = self.fct_frequency(numusers=len(users))

        for j, user in enumerate(users):
            if j == 0:
                prob_user = copy.deepcopy(user.presence)
            else:
                prob_user += user.presence

        prob_user = normalize(prob_user).values
        j = len(users)

        prob_joint = normalize(prob_user * prob_usage)

        pattern = self.fct_duration_pattern()
        duration = len(pattern)

        for i in range(freq):

            u = np.random.random()
            start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
            end = start + duration

            if end > (24 * 60 * 60):
                end = 24 * 60 * 60
            difference = end - start
            consumption[start:end, j, ind_enduse] = pattern[:difference]

            prob_user[start:end] = 0
            prob_user = normalize(prob_user)

            prob_usage[start - duration:end] = 0
            prob_usage = normalize(prob_usage)

        return consumption


class Wc(EndUse):

    def __init__(self, name='Wc', **kwargs):
        super(Wc, self).__init__(**kwargs)
        self.name = name

    def fct_frequency(self, age=None, gender=None):
        f_stats = self.statistics['frequency']
        distribution = getattr(np.random, f_stats['distribution'].lower())

        average = f_stats['average'][age][gender]

        return distribution(average)

    def fct_duration_intensity(self):

        flush_interuption = self.statistics['subtype'][self.name]['flush_interuption']
        prob_flush_interuption = self.statistics['prob_flush_interuption']

        intensity = self.statistics['intensity']

        average = self.statistics['subtype'][self.name]['duration']

        # dist = duration_decorator(getattr(np.random, d_stats['distribution'].lower()))

        average = to_timedelta(np.log(to_timedelta(average).total_seconds()) - 0.5)

        # add water savings option
        if flush_interuption:
            v = np.random.random() * 100
            if v < prob_flush_interuption:
                average /= 2.0

        duration = int(average.total_seconds())

        return duration, intensity


    def simulate(self, consumption, users=None, ind_enduse=None):

        prob_usage = self.usage_probability().values

        for j, user in enumerate(users):
            freq = self.fct_frequency(age=user.age, gender=user.gender)
            prob_user = user.presence.values

            for i in range(freq):
                check = False

                duration, intensity = self.fct_duration_intensity()

                prob_joint = normalize(prob_user * prob_usage)

                # get rid of overlaps
                for tries in range(10):
                    u = np.random.uniform()
                    start = np.argmin(np.abs(np.cumsum(prob_joint) - u))
                    end = start + duration

                    if np.sum(consumption[start:end, j, ind_enduse]) == 0:
                        check = True
                        break
                    else:
                        print('consumption overlap')
                # print(start, end)
                # print(consumption.time)
                if check:
                    prob_user[start:end] = 0
                    prob_user = normalize(prob_user)

                    prob_usage[start:end] = 0
                    prob_usage = normalize(prob_usage)

                    consumption[start:end, j, ind_enduse] = intensity

        return consumption

class WcNormal(Wc):

    def __init__(self, name='WcNormal', **kwargs):
        super(WcNormal, self).__init__(**kwargs)
        self.name = name


class WcNormalSave(Wc):

    def __init__(self, name='WcNormalSave', **kwargs):
        super(WcNormalSave, self).__init__(**kwargs)
        self.name = name


class WcNew(Wc):

    def __init__(self, name='WcNew', **kwargs):
        super(WcNew, self).__init__(**kwargs)
        self.name = name


class WcNewSave(Wc):

    def __init__(self, name='WcNewSave', **kwargs):
        super(WcNewSave, self).__init__(**kwargs)
        self.name = name



import copy
import numpy as np
import scipy.stats as sstats
import pandas as pd
from pysimdeum.core.utils import Base
from dataclasses import dataclass, field
from typing import Any, Callable, Literal
from pysimdeum.core.statistics import Statistics


@dataclass
class Presence:
    """Class representing the presence and the water use activity of users in a house."""

    weekday: bool
    user: Any
    stats: Statistics

    up: pd.Timedelta = field(init=False)
    go: pd.Timedelta = field(init=False)
    home: pd.Timedelta = field(init=False)
    sleep: pd.Timedelta = field(init=False)

    _prob_getting_up: Any = field(init=False, repr=False)
    _prob_leaving_house: Any = field(init=False, repr=False)
    _prob_being_away: Any = field(init=False, repr=False)
    _prob_sleep: Any = field(init=False, repr=False)

    def __post_init__(self) -> None:

        if self.weekday:
            diurnal = copy.deepcopy(self.stats.diurnal_pattern[self.user.age])
        else:
            diurnal = copy.deepcopy(self.stats.diurnal_pattern['weekend'])

        for key, val in diurnal.items():
            dist = val['dist']
            del val['dist']
            # dist = getattr(pm, dist)
            dist = getattr(sstats, dist)
            newval = dict()
            translate = {'mu': 'loc',
                         'sd': 'scale'}

            for x, y in val.items():
                newval[translate[x]] = round(pd.Timedelta(y).total_seconds() / 60)

            # setattr(self, '_prob_' + key, dist.dist(**newval))
            setattr(self, '_prob_' + key, dist(**newval))

        self.up = self.sample_single_property('_prob_getting_up')

        self.sleep = self.up - self.sample_single_property('_prob_sleep') + pd.Timedelta(days=1)

        self.go = self.sample_single_property('_prob_leaving_house')

        if self.go < self.up:
            self.go = self.up + pd.Timedelta(minutes=30)

        self.home = self.go + self.sample_single_property('_prob_being_away')

        if self.home < self.go:
            self.home = self.go  # actually no leave

        if self.sleep < self.home:
            self.home = self.sleep - pd.Timedelta(minutes=30)

    def print(self) -> None:
        """Method to print the main properties of the user's presence"""

        print('up:', self.up)
        print('go:', self.go)
        print('home:', self.home)
        print('sleep:', self.sleep)

    def sample_single_property(self, prop: str) -> pd.Timedelta:
        """Function to draw random time values from the single time properties (e.g., getting up, leave house, ...) of the users"""

        prob_fct = getattr(self, prop)
        x = prob_fct.rvs()
        x = int(np.round(x))
        x = pd.Timedelta(minutes=x)
        return x

    @staticmethod
    def timestamp2str(x: pd.Timedelta) -> str:
        """Transform a Python Pandas Timedelta object to a string in the format HH:MM"""
        
        return str(x.components.hours).zfill(2) + ':' + str(x.components.minutes).zfill(2) # + ':' + str(
        # x.components.seconds).zfill(2)

    def timeindexer(self, l, value, a, b):

        if a < b:
            l[a:b] = value
        else:
            l[a:len(l)] = value
            l[0:b] = value
        return l

    def pdf(self, peak=0.65, normal=0.335, away=0.0, night=0.015):
        index = pd.timedelta_range(start='00:00:00', end='24:00:00', freq='1Min')
        pdf = pd.Series(index=index, dtype='float64')

        up = int((self.up.total_seconds()) / 60) % 1440
        up_p30 = int((up + 30)) % 1440

        go = int(self.go.total_seconds() / 60) % 1440
        go_m30 = int(go - 30) % 1440

        home = int(self.home.total_seconds() / 60) % 1440
        home_p30 = int(home + 30) % 1440

        sleep = int(self.sleep.total_seconds() / 60) % 1440
        sleep_m30 = int(sleep - 30) % 1440

        pdf = self.timeindexer(pdf, 'normal', up_p30, go_m30)
        pdf = self.timeindexer(pdf, 'normal', home_p30, sleep_m30)
        pdf = self.timeindexer(pdf, 'peak', up, up_p30)
        pdf = self.timeindexer(pdf, 'peak', go_m30, go)
        pdf = self.timeindexer(pdf, 'peak', home, home_p30)
        pdf = self.timeindexer(pdf, 'peak', sleep_m30, sleep)
        pdf = self.timeindexer(pdf, 'night', sleep, up)
        pdf = self.timeindexer(pdf, 'away', go, home)

        cnts = pdf.value_counts(normalize=True)
        try:
            cnts = cnts.drop('away')
        except:
            pass
        cnts /= cnts.sum()
        try:
            pdf[pdf == 'peak'] = peak / cnts['peak']
        except:
            pass
        try:
            pdf[pdf == 'normal'] = normal / cnts['normal']
        except:
            pass
        try:
            pdf[pdf == 'night'] = night / cnts['night']
        except:
            pass
        try:
            pdf[pdf == 'away'] = 0.0
        except:
            pass


        pdf = pdf.astype('float').resample('1S').fillna('ffill')[:-1]

        pdf /= np.sum(pdf)  # normalize

        # transform timedeltas to strings for indexing later on
        return pdf

# removed reference to house. house user belongs to house and not also vice versa


@dataclass
class User(Base):

    gender: Literal['male', 'female'] = None
    age: Literal['child', 'teen', 'adult', 'home_ad', 'work_ad', 'senior'] = None  #TODO: Detangle home and work adult from adult
    job: bool = True

    presence: Presence = field(init=False, repr=False)

    def __post_init__(self):

        if self.age == 'adult':
            if self.job:
                self.age = 'work_ad'
            else:
                self.age = 'home_ad'


    def compute_presence(self, weekday=True, statistics=None, peak=0.65, normal=0.335, away=0.0, night=0.15):

        presence = Presence(user=self, weekday=weekday, stats=statistics)
        pdf = presence.pdf(peak=peak, normal=normal, away=away, night=night)
        self.presence = pdf

        return self.presence



import pandas as pd
from matplotlib import pyplot as plt

def dishwasher_daily_pattern(x=None, resolution='1s'):

    # todo: last hour of day is not talking into account in the pattern

    if x is None:
        x = '70 49 33 69 30 19 9 37 80 52 52 45 53 78 34 44 38 101 489 390 170 185 240 574 70'
    data = list(map(float, x.split(' ')))
    index = pd.timedelta_range(start='00:00:00', freq='1H', periods=len(data))
    s = pd.Series(data=data, index=index)
    s = s.resample(resolution).mean().interpolate(method='linear') # todo interpolate with cubic splines does not work
    s = s[s.index.days == s.index[0].days]


    return s

def dishwasher_enduse_pattern(resolution='1s', periods=7200):

    value = 1/6

    index = pd.timedelta_range(start='00:00:00', freq=resolution, periods=periods)
    s = pd.Series(0, index=index)

    s.iloc[0:24] = value
    s.iloc[1860:1884] = value
    s.iloc[3660:3684] = value
    s.iloc[5460:5472] = value

    # s.index = s.index - s.index[0]

    return s

if __name__ == '__main__':

    pass

    # x = '70 49 33 69 30 19 9 37 80 52 52 45 53 78 34 44 38 101 489 390 170 185 240 574'
    # data = list(map(float, x.split(' ')))
    # index = pd.DatetimeIndex(start='2018/1/1', freq='1H', periods=24)
    # s1 = pd.Series(data=data, index=index)
    #
    # s2 = make_daily_pattern(x=x)
    #
    # fig = plt.figure(1)
    # ax = fig.add_subplot(111)
    #
    # s1.plot(ax=ax, marker='o', linestyle='None')
    # s2.plot(ax=ax)
    # plt.show()




import pandas as pd
from matplotlib import pyplot as plt


def washingmachine_daily_pattern(x=None, resolution='1s'):

    if x is None:
        x = '34 24 17 43 41 57 57 212 290 338 392 325 226 222 200 175 124 106 139 185 133 165 101 147 34'
    data = list(map(float, x.split(' ')))
    index = pd.timedelta_range(start='00:00:00', freq='1H', periods=25)
    s = pd.Series(data=data, index=index)
    s = s.resample(resolution).mean().interpolate(method='linear')
    s = s[s.index.days == s.index[0].days]
    return s


def washingmachine_enduse_pattern(resolution='1s', periods=7200):

    value = 1/6

    index = pd.timedelta_range(start='00:00:00', freq=resolution, periods=periods)
    s = pd.Series(0, index=index)

    s.iloc[0:121] = value
    s.iloc[3600:3660] = value
    s.iloc[4920:4980] = value
    s.iloc[6120:6180] = value

    # s.index = s.index - s.index[0]

    return s


